use std::{
    io::{self, Write},
    str::FromStr,
};

use crate::{
    lexer::{self, Lexer, Source, Token},
    parser::Parser,
};

struct GlobalArgs {
    subcommand: Option<Subcommand>,
}

enum Subcommand {
    Repl,
    Help,
}

#[derive(Debug)]
pub enum ParseArgsError {
    InvalidSubcommand(String),
    UnexpectedArgument(String),
}
impl std::error::Error for ParseArgsError {}
impl std::fmt::Display for ParseArgsError {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        match self {
            ParseArgsError::InvalidSubcommand(c) => write!(f, "invalid subcommand: {}", c),
            ParseArgsError::UnexpectedArgument(c) => write!(f, "unexpected argument: {}", c),
        }
    }
}

impl FromStr for Subcommand {
    type Err = ParseArgsError;

    fn from_str(input: &str) -> Result<Subcommand, Self::Err> {
        // TODO: errors should show similar commands?
        match input {
            "help" => Ok(Subcommand::Help),
            "repl" => Ok(Subcommand::Repl),
            _ => Err(ParseArgsError::InvalidSubcommand(input.to_string())),
        }
    }
}

fn usage(exit_code: i32) -> ! {
    println!("Hang in there! TODO: Implement help!");
    std::process::exit(exit_code);
}

fn parse_global_args() -> Result<GlobalArgs, ParseArgsError> {
    let mut subcommand = None;
    let mut args = std::env::args().skip(1);

    while let Some(arg) = args.next() {
        match arg.as_str() {
            "--help" | "-h" => {
                usage(0);
            }
            s if subcommand.is_none() => {
                subcommand = Some(s.parse()?);
            }
            _ => return Err(ParseArgsError::UnexpectedArgument(arg.to_string())),
        }
    }

    return Ok(GlobalArgs { subcommand });
}

fn repl() -> ! {
    println!("lytlang repl (^D to stop)");
    print!("> ");
    std::io::stdout().flush().expect("failed to flush stdout");
    let mut iter = std::io::stdin().lines().enumerate().into_iter();
    while let Some((i, l)) = iter.next() {
        let s = match &l {
            Ok(s) => s.as_str(),
            Err(e) => {
                eprintln!("error reading line from stdin: {}", e);
                continue;
            }
        };
        let tokens = match Lexer::new(s, Source::Repl(i)).collect() {
            Ok(tokens) => tokens,
            Err(e) => {
                eprintln!("error lexing input: {:?}", e);
                continue;
            }
        };
        match Parser::default().parse(tokens) {
            Ok(expr) => println!("< {:?}", expr),
            Err(e) => {
                eprintln!("error parsing tokens: {:?}", e);
                continue;
            }
        }
        println!("> ");
        std::io::stdout().flush().expect("failed to flush stdout");
    }
    std::process::exit(0);
}

pub fn run() -> crate::Result<()> {
    match parse_global_args()?.subcommand {
        Some(Subcommand::Repl) => repl(),
        Some(Subcommand::Help) => usage(0),
        None => {
            println!("no subcommand specified");
            usage(1)
        }
    }
}
