use std::{iter::Peekable, vec::IntoIter};

use crate::{ast::*, lexer};

pub struct Parser {
    num_tokens_parsed: usize,
}

impl Default for Parser {
    fn default() -> Self {
        Self {
            num_tokens_parsed: 0,
        }
    }
}

#[derive(Debug, PartialEq)]
pub enum Error {
    Lexer(lexer::Error),
    Unexpected(lexer::Token),
}

impl<'a> From<lexer::Error> for Error {
    fn from(value: lexer::Error) -> Self {
        Error::Lexer(value)
    }
}

impl Parser {
    pub fn parse_str(&mut self, code: &'static str) -> Result<Expression, Error> {
        let lexer = lexer::Lexer::new(code, lexer::Source::Unknown);
        let tokens = lexer.collect::<Result<Vec<lexer::Token>, lexer::Error>>()?;
        Ok(self.parse(tokens)?)
    }

    pub fn parse(&mut self, tokens: Vec<lexer::Token>) -> Result<Expression, Error> {
        let iter = tokens.into_iter().peekable();
        self.expression(iter)
    }

    fn expression(
        &mut self,
        tokens: Peekable<IntoIter<lexer::Token>>,
    ) -> Result<Expression, Error> {
        Ok(Expression::Unit)
    }

    pub fn num_tokens_parsed(&self) -> usize {
        self.num_tokens_parsed
    }
}

#[cfg(test)]
mod test {
    use super::*;

    #[test]
    fn addition_operation() {
        let mut parser = Parser::default();
        assert_eq!(
            parser.parse_str("3 + 9"),
            Ok(Expression::InfixOperation(
                Box::new(Expression::Integer(3)),
                Operator::Add,
                Box::new(Expression::Integer(9))
            ))
        );
        assert_eq!(parser.num_tokens_parsed(), 3);
    }
}
